/*
 * Copyright (c) 1997-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2004      Abhishek Rai,
 * Copyright (c) 2005      Sunil Satnur,
 * Copyright (c) 2006      Avishay Traeger,
 * Copyright (c) 2001-2006 Stony Brook University,
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: mmap.c,v 1.110 2006/12/13 17:00:21 kolya Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "raif.h"
#include "loadbal.h"
#include "ECC/levels.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define PageUptodate(page) Page_Uptodate(page)
#define unlock_page(page) UnlockPage(page)
#define wait_on_page_locked(page) wait_on_page(page)
#endif

STATIC page_t *raif_read_cache_page(file_t *file, inode_t *hidden_inode, int bindex, unsigned page_index);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int raif_do_writepage(page_t * page, inode_t * inode, unsigned bindex, unsigned hidden_page_index)
#else
int raif_do_writepage(page_t * page, inode_t * inode, unsigned bindex, unsigned hidden_page_index, struct writeback_control *wbc)
#endif
{
	int i, err = 0;
	inode_t *lower_inode;
	page_t *lower_page;
	inode_t ** hidden_inodes = NULL;
	page_t ** hidden_pages = NULL;
	unsigned sbr, level, iECCw, iECC, bnum;

#ifdef RAIF6_ROW_DIAG
	int diag1_page_index = -1, diag2_page_index = -1;
#endif

	CALCULATE_ECC_BRANCHES (inode, hidden_page_index, &level, &iECCw, &sbr, &iECC, &bnum);

#ifdef RAIF6_ROW_DIAG

	if (iECCw == 2) {
		CALCULATE_DIAG_PAGE_LOCATIONS_RDP(inode, hidden_page_index, bindex, &diag1_page_index, &diag2_page_index );
	}

#endif

	hidden_inodes = (inode_t**)KMALLOC(sizeof(inode_t*) * bnum, GFP_KERNEL);
	hidden_pages  = (page_t**)KMALLOC(sizeof(page_t*) * bnum, GFP_KERNEL);
	if (!hidden_pages || !hidden_inodes) {
		err = -ENOMEM;
		goto out;
	}

	for (i = 0; i < bnum ;i++) {
		hidden_inodes[i] = NULL;
		hidden_pages[i] = NULL;
	}

	hidden_inodes[0] = itohi_index(inode, bindex);
	if (hidden_inodes[0] == NULL) {
		err = -EIO;
		goto out;
	}

#ifdef RAIF6_ROW_DIAG

	if (iECCw == 2) {
		hidden_inodes[1] = itohi_index(inode, bnum - 2);
		hidden_inodes[2] = itohi_index(inode, bnum - 1);
		if (!hidden_inodes[1] || !hidden_inodes[2]) {
			err = -EIO;
			goto out;
		}
		
		hidden_inodes[3] = hidden_inodes[2];

	} else 

#endif

	for (i = 0 ; i < iECCw ; i++) {
		hidden_inodes[i + 1] = itohi_index(inode, (i + iECC) % bnum);
		if (hidden_inodes[i + 1] == NULL) {
			err = -EIO;
			goto out;
		}
	}

	/* We need the old unmodified page only in the case when we are calculating parity
	 * i.e. when iECCw >= 1. So, we call raif_read_cache_page only when iECCw >= 1.
	 */

	if (iECCw < 1) {
		hidden_pages[0] = grab_cache_page(hidden_inodes[0]->i_mapping, hidden_page_index);
		if (!hidden_pages[0]) {
			err = -ENOMEM;
			goto out;
		}
	} else {

		hidden_pages[0] = raif_read_cache_page(NULL, hidden_inodes[0], bindex, hidden_page_index);

		if (IS_ERR(hidden_pages[0])) {
			err = PTR_ERR(hidden_pages[0]);
			goto out; 
		}

		lock_page(hidden_pages[0]);
	}

	kmap(hidden_pages[0]);

#ifdef RAIF6_ROW_DIAG

	if (iECCw == 2) {
		hidden_pages[1]= raif_read_cache_page(NULL, hidden_inodes[1], bnum - 2, hidden_page_index);

		if (IS_ERR(hidden_pages[1])) {
			err = PTR_ERR(hidden_pages[1]);
			i = 0;
			goto undo_pages; 
		}

		lock_page(hidden_pages[1]);
		kmap(hidden_pages[1]);

		if (diag1_page_index != -1) {

			hidden_pages[2] = raif_read_cache_page(NULL, hidden_inodes[2], bnum - 1, diag1_page_index);

			if (IS_ERR(hidden_pages[2])) {
				err = PTR_ERR(hidden_pages[2]); 
				i = 1;
				goto undo_pages; 
			}
			lock_page(hidden_pages[2]);
			kmap(hidden_pages[2]);

		}

		if (diag2_page_index != -1) {
			hidden_pages[3] = raif_read_cache_page(NULL, hidden_inodes[3], bnum - 1, diag2_page_index);

			if (IS_ERR(hidden_pages[3])) {
				err = PTR_ERR(hidden_pages[3]);
				i = 2;
				goto undo_pages;
			}

			lock_page(hidden_pages[3]);
			kmap(hidden_pages[3]);
		}
	} else 

#endif

	for (i = 0; i < iECCw; i++) {
		hidden_pages[i + 1] = raif_read_cache_page(NULL, hidden_inodes[i + 1], (i + iECC) % bnum, hidden_page_index);

		if (IS_ERR(hidden_pages[i + 1])) {
			err = PTR_ERR(hidden_pages[i + 1]); 

#ifdef RAIF6_ROW_DIAG
undo_pages:
#endif
			for (; i >= 0; i--) {
				if (hidden_pages[i]) {
					kunmap(hidden_pages[i]);
					unlock_page(hidden_pages[i]);
					page_cache_release(hidden_pages[i]);
				}
			}
			goto out;
		}

		lock_page(hidden_pages[i + 1]);
		kmap(hidden_pages[i + 1]);
	}

	kmap(page);

#ifdef RAIF6_ROW_DIAG
	if (iECCw == 2) {
		update_ECC_RDP(hidden_pages, page_address(page), 0, PAGE_CACHE_SIZE, 0);
	} else
#endif

	update_ECC(hidden_pages, page_address(page), 0, 0, PAGE_CACHE_SIZE, iECCw + 1, iECCw, 0);

	kunmap(page);

#ifdef RAIF6_ROW_DIAG
	/* iECCw + 1 is the number times the loop below is executed.
	 * For row-diagonal parity, the number of hidden_pages is 4.
	 * So, we set iECCw to 3. For other algorithms, it is usually iECCw+1
	 */
	if (iECCw == 2)
		iECCw = 3;
#endif

	for (i = 0; i < iECCw + 1; i++) {
		if (hidden_pages[i])
			kunmap(hidden_pages[i]);
	}

	for (i = 0; i < iECCw + 1; i++) {
		lower_inode = hidden_inodes[i];
		lower_page = hidden_pages[i];

		if (lower_page == NULL)
			continue;

		spin_lock(&(lower_inode->i_mapping->private_lock));

		if (page_has_buffers(lower_page)) {
			struct buffer_head *head = page_buffers(lower_page);
			struct buffer_head *bh = head;

			do {
				set_buffer_dirty(bh);
				bh = bh->b_this_page;
			} while (bh != head);
		}

		spin_unlock(&(lower_inode->i_mapping->private_lock));

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		if (wbc->sync_mode != WB_SYNC_NONE)
			wait_on_page_writeback(lower_page);
#endif

		if (PageWriteback(lower_page)) {
			unlock_page(lower_page);
			page_cache_release(lower_page);
			continue;
		}

		/* We cache all the parity pages in the lower FS and all non-parity pages in
		 * upper FS. So, we call SetPageReclaim only for non-parity pages.
		 */

		if (i == 0)
			SetPageReclaim(lower_page);

		/* call lower writepage (expects locked page) */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		err = lower_inode->i_mapping->a_ops->writepage(lower_page);
#else
		err = lower_inode->i_mapping->a_ops->writepage(lower_page, wbc);
#endif
		if (err) {
			if (i == 0)
				ClearPageReclaim(lower_page);

			page_cache_release(lower_page);

			for (i++; i < iECCw + 1; i++) {
				unlock_page(hidden_pages[i]);
				page_cache_release (hidden_pages[i]);
			}

			break;
		}
		if ( (i == 0) && !PageWriteback(lower_page)) {
			ClearPageReclaim(lower_page);
		}

		page_cache_release(lower_page); /*because read_cache_page increased refcnt*/
	}

out:
	if (hidden_inodes)
		KFREE (hidden_inodes);
	if (hidden_pages)
		KFREE (hidden_pages);
	return err;

}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
STATIC int raif_writepage(page_t *page)
#else
STATIC int raif_writepage(page_t *page, struct writeback_control *wbc)
#endif
{
	int err = -EIO;
	inode_t *inode;
	unsigned bindex, bnum, page_index, lower_page_index;

	print_entry_location();

	CALCULATE_PAGE_LOCATION_WRITE (page, &inode, &bnum, &page_index, &bindex, &lower_page_index);

	for (;bindex < bnum; bindex++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		err = raif_do_writepage(page, inode, bindex, lower_page_index);
#else
		err = raif_do_writepage(page, inode, bindex, lower_page_index, wbc);
#endif
		if (err)
			break;
	}

	if (err)
		ClearPageUptodate(page);
	else
		SetPageUptodate(page);

	unlock_page(page);
	return err;
}


STATIC page_t *raif_read_cache_page(file_t *file, inode_t *hidden_inode, int bindex, unsigned page_index)
{
	page_t *hidden_page;
	int err;
/*#ifdef RAIF_LB
        struct timeval b4, af;
#endif*/

	/*LB_UPDATE_PREPARE;*/
	hidden_page = read_cache_page(hidden_inode->i_mapping, page_index, 
			(filler_t *) hidden_inode->i_mapping->a_ops->readpage,
			(void *) file);

	if (IS_ERR(hidden_page)) 
		goto out;

	/* wait for the page data to show up
 	 * (signaled by readpage as unlocking the page)
	 */
	wait_on_page_locked(hidden_page);

	if (!PageUptodate(hidden_page)) {
		/* call readpage() again if we returned from wait_on_page with a
 		 * page that's not up-to-date; that can happen when a partial
 		 * page has a few buffers which are ok, but not the whole
 		 * page. 
		 */
		lock_page(hidden_page);
		err = hidden_inode->i_mapping->a_ops->readpage(file, hidden_page);
		if (err) {
			page_cache_release(hidden_page);
			hidden_page = ERR_PTR(err);
			goto out;
		}
		wait_on_page_locked(hidden_page);
		if (!PageUptodate(hidden_page)) {
			page_cache_release(hidden_page);
			err = -EIO;
			hidden_page = ERR_PTR(err);
			goto out;
		}
	}
	/*LB_UPDATE_COMMIT_F(file, bindex, read);*/

out:
	return hidden_page;
}

#ifdef RAIF6_ROW_DIAG
/* Recovery function for the Row-Diagonal Parity Algorithm.
 * The Row-Diagonal Parity Algorithm has to be handled `specially' in many places because it uses 
 * data from many stripes rather than just one stripe.
 *
 * For recovery, we construct the RDP array at pass it to the recover_ECC_RDP function.
 *
 * This can be made more efficient.
 */
STATIC int raif_recover_page_rdp(file_t *file, inode_t *inode, char *page_data, int bindex1, unsigned hidden_page_index)
{
	int err=0;
	page_t *** arrayPages = NULL;
	page_t ** arrayData = NULL;
	int * flagData = NULL;
	int ** flags = NULL;
	page_t * page;
	int i, j;
	unsigned bl_num, bnum, sus, sush, strp, strp_inb, iECCw, imax, sbr;
	inode_t * hidden_inode;
	char * hidden_page_data;
	loff_t min_pages, real_size;
	int changed_bindex1 = -1, changed_bindex2 = -1;

	int bindex2 = -1;


	bnum = ibnum(inode);
	sush = STRIPING_UNIT_SHIFT(inode);

	sus = STRIPING_UNIT_SIZE(inode);

	sbr = START_BRANCH(inode);

	iECCw = ECCW(LEVEL(inode));

	imax = bnum - iECCw;

	hidden_page_index = DEC_IF_TAG(hidden_page_index);

	strp = hidden_page_index >> sush;

	bl_num = strp /(bnum - 2);
	strp_inb = strp % (bnum - 2);

	min_pages = bl_num * sus * (bnum - 2);

	flagData = KMALLOC(sizeof(int) * (bnum - 2) * bnum * sus, GFP_KERNEL);

	if(!flagData){
		err = -ENOMEM;
		goto out;
	}

	flags = KMALLOC(sizeof(int *) * (bnum - 2) * sus, GFP_KERNEL);

	if(!flags){
		err = -ENOMEM;
		goto out;
	}

	arrayData = KMALLOC(sizeof(page_t *) * (bnum-2) * bnum * sus, GFP_KERNEL);
	
	if(!arrayData){
		err = -ENOMEM;
		goto out;
	}

	arrayPages = KMALLOC(sizeof(page_t **) * (bnum-2) * sus, GFP_KERNEL);

	if(!arrayPages){
		err = -ENOMEM;
		goto out;
	}

	for(i = 0;i < (bnum - 2) * sus;i++){
		arrayPages[i] = &arrayData[i*bnum];
		flags[i] = &flagData[i*bnum];
		for(j = 0; j < bnum ;j++){
			arrayPages[i][j] = NULL;
			flags[i][j] = 0;
		}
	}
	

	for(j = 0;j < bnum;j++){
		loff_t i_size, i_pages;
		int lcount;
		int changed_bindex; 
		
		if(j >= imax)
			changed_bindex = j;
		else
			changed_bindex = (j + imax - sbr % imax) % imax;
		if((j == bindex1) || (j == bindex2))
			continue;
		hidden_inode = itohi_index(inode, j);
		i_size = hidden_inode->i_size;


		i_pages = i_size >> PAGE_CACHE_SHIFT;
		if(i_size % PAGE_CACHE_SIZE)
			i_pages++;
		if(i_pages <= min_pages)
			continue;
		
		if((i_pages - min_pages) > (bnum - 2) * sus)
			lcount = (bnum - 2) * sus;
		else
			lcount = i_pages - min_pages;

		for(i = 0;i < lcount;i++){
			page_t * temp;
			temp=raif_read_cache_page(file, hidden_inode, j, INC_IF_TAG(min_pages + i));
			if(IS_ERR(temp)){
				if(bindex2 == -1){
					bindex2 = j;
					break;
				}
				else if(bindex2 != j){
					err = PTR_ERR(temp);
					goto out_clear_array;	
				}
				
				continue;
			}
			arrayPages[i][changed_bindex] = temp;
			flags[i][changed_bindex] = 1;
		}
	}

	if(bindex1 != -1 && bindex1 < imax)
		changed_bindex1 = (bindex1 + imax - sbr % imax) % imax;
	else
		changed_bindex1 = bindex1;

	if(bindex2 != -1 && bindex2 < imax)
		changed_bindex2 = (bindex2 + imax - sbr % imax) % imax;
	else
		changed_bindex2 = bindex2;

	if(changed_bindex1 != -1){
		for(i = 0;i < (bnum - 2) * sus;i++){
			if(arrayPages[i][changed_bindex1]){
				page_cache_release(arrayPages[i][changed_bindex1]);
				arrayPages[i][changed_bindex1] = NULL;
			}
		}
	}

	if(changed_bindex2 != -1){
		for(i = 0;i < (bnum - 2) * sus;i++){
			if(arrayPages[i][changed_bindex2]){
				page_cache_release(arrayPages[i][changed_bindex2]);
				arrayPages[i][changed_bindex2] = NULL;
			}
		}
	}

	err = recover_ECC_RDP(arrayPages, bnum, changed_bindex1, changed_bindex2, sus, hidden_page_index - min_pages);

	if(err)
		goto out_clear_array;

	page = arrayPages[hidden_page_index - min_pages][changed_bindex1];


	hidden_page_data = (char *)kmap(page);
	/* Copy the recovered page to page_data */
	hidden_inode = itohi_index(inode, bindex1);
	real_size = hidden_inode->i_size - (INC_IF_TAG(hidden_page_index) << PAGE_CACHE_SHIFT);
	if (real_size <= 0)
		memset(page_data, 0, PAGE_CACHE_SIZE);
	else if (real_size < PAGE_CACHE_SIZE) {
		memcpy(page_data, hidden_page_data, real_size);
		memset(page_data + real_size, 0, PAGE_CACHE_SIZE - real_size);
	} else {
		memcpy(page_data, hidden_page_data, PAGE_CACHE_SIZE);
	}
	kunmap(page);


out_clear_array:

	for(i = 0;i < (bnum - 2) * sus;i++){
		for(j = 0;j < bnum;j++){
			if(arrayPages[i][j]){
				if(!flags[i][j])
					free_page((unsigned long) arrayPages[i][j]);
				else
					page_cache_release(arrayPages[i][j]);
			}
		}
	}

out:
	if(flagData)
		KFREE(flagData);

	if(flags)
		KFREE(flags);

	if(arrayData)
		KFREE(arrayData);

	if(arrayPages)
		KFREE(arrayPages);

	return err;
}
#endif

STATIC int raif_recover_page(file_t *file, inode_t *inode, char *page_data, int bindex, unsigned hidden_page_index)
{
	struct {
		page_t **pagelist;
		int numPages;
		int numBadPages;
		int *badPages;
	} recover_list;
	inode_t *hidden_inode;
	page_t *page=NULL;
	char *hidden_page_data;
	int err, bnum, npages=0, i, level; 
	int su_index, level_start_index, real_size;
	unsigned sus, imax, iECCw;

#ifdef RAIF6_ROW_DIAG
	if(LEVEL(inode) == 6)
		return raif_recover_page_rdp(file, inode, page_data, bindex, hidden_page_index);
#endif

	recover_list.pagelist = NULL;
	recover_list.badPages = NULL;

	hidden_inode    = itohi_index(inode, bindex);

	// XXX "This code should use macros from ECC.h"
	bnum = ibnum(inode);
	iECCw = ECCW(LEVEL(inode));
	sus = STRIPING_UNIT_SIZE(inode);

	switch(LEVEL(inode)) {
		case 0:
			/* Cant recover in RAIF 0 */
			err = -EIO; 
			goto out_return;
		case 1:
			/* return the first successful mirrored page */
			for (i = 0; i < bnum; i++) {
				if (i == bindex) 
					continue; /* No point in reading from failed branch */

				hidden_inode = itohi_index(inode, i);
				page = raif_read_cache_page(file, hidden_inode, i, hidden_page_index); 
				if (!IS_ERR(page)) {
					goto out_copy;
				}
			}
			err = -EIO;
			goto out_return;

		case 4:
		case 5:
			err = -ENOMEM;
			recover_list.pagelist = NULL;
			recover_list.pagelist = KMALLOC(sizeof(page_t *) * bnum, GFP_KERNEL);
			if (!recover_list.pagelist)
				goto out_return;
			recover_list.badPages = NULL;
			recover_list.badPages = KMALLOC(sizeof(int) * bnum, GFP_KERNEL);
			if (!recover_list.badPages)
				goto out_free_pages;

			recover_list.numPages = bnum;
			recover_list.numBadPages = 0;
			imax = bnum - iECCw;
			level = hidden_page_index / (sus * imax);
			su_index = hidden_page_index % sus;
			level_start_index = imax * sus * level;

			/* Read all the pages in this stripe */
			for (i=0; i < bnum; i++) {
				if (i == bindex) {
alloc_page:			page = alloc_pages(GFP_KERNEL|__GFP_REPEAT|__GFP_ZERO, 0);
					if (IS_ERR(page)) {
						err = -ENOMEM;
						npages = i;
						goto out_free_pages;
					}
					recover_list.badPages[recover_list.numBadPages++] = i;
					recover_list.pagelist[i] = page;
					continue;
				}
				hidden_inode    = itohi_index(inode, i);

				page = raif_read_cache_page(file, 
					hidden_inode, i,
					hidden_page_index);
				if (IS_ERR(page)) {
					goto alloc_page;
				}
				recover_list.pagelist[i] = page;
			}
			npages = bnum;

			err = recover_ECC(recover_list.pagelist, 
					recover_list.numPages,
					recover_list.badPages, 
					recover_list.numBadPages, iECCw);
			if (err)
				goto out_free_pages;
			page = recover_list.pagelist[bindex];

			break;
		
		default: 
			err = -EIO;
			goto out_return;
	}

out_copy:
	switch(LEVEL(inode)) {
		case 1:
		case 4:
		case 5:
			err = 0;
			hidden_page_data = (char *)kmap(page);
			/* Copy the recovered page to page_data */
			hidden_inode = itohi_index(inode, bindex);
			real_size = hidden_inode->i_size - (hidden_page_index << PAGE_CACHE_SHIFT);
			if (real_size <= 0)
				memset(page_data, 0, PAGE_CACHE_SIZE);
			else if (real_size < PAGE_CACHE_SIZE) {
				memcpy(page_data, hidden_page_data, real_size);
				memset(page_data + real_size, 0, PAGE_CACHE_SIZE - real_size);
			} else {
				memcpy(page_data, hidden_page_data, PAGE_CACHE_SIZE);
			}
			kunmap(page);
			break;

		default:
			err = -EIO;
			break;
	}

out_free_pages:
	switch(LEVEL(inode)) {
		case 1:
			free_page((unsigned long)page);
			break;

		case 4:
		case 5:
			for (i = 0; i < npages; i++) 
				free_page((unsigned long)recover_list.pagelist[i]);

			if (recover_list.pagelist)
				KFREE(recover_list.pagelist);
			if (recover_list.badPages)
				KFREE(recover_list.badPages);
			break;

		default:
			err = -EIO;
	}
out_return:
	return err;
}

STATIC int raif_do_readpage(file_t *file, char *page_data, int bindex, unsigned hidden_page_index)
{
	int err = -EIO;
	dentry_t *dentry;
	file_t *hidden_file = NULL;
	dentry_t *hidden_dentry;
	inode_t *inode;
	inode_t *hidden_inode;
	page_t *hidden_page;
	char *hidden_page_data;
	int real_size;

	print_entry_location();

	dentry = file->f_dentry;
	inode = dentry->d_inode;

	if (ftopd(file) != NULL)
		hidden_file = ftohf_index(file, bindex);
	hidden_dentry   = dtohd_index(dentry, bindex);
	hidden_inode    = itohi_index(inode, bindex);

	
	CURRENT_HIDDEN_PAGE = raif_read_cache_page(hidden_file, hidden_inode,
						bindex, hidden_page_index);

	if (IS_ERR(CURRENT_HIDDEN_PAGE)) {
		err = raif_recover_page(file, inode, page_data, bindex, hidden_page_index);
		goto out_release;
	}

	CURRENT_HIDDEN_PAGEDATA = (char *) kmap(CURRENT_HIDDEN_PAGE);

	err = 0;

	//real_size = inode->i_size - (page_index << PAGE_CACHE_SHIFT);
	real_size = hidden_inode->i_size - (hidden_page_index << PAGE_CACHE_SHIFT);
	fist_dprint(7, "%s: real_size=%d. i_size=%d(%d). i_blocks=%d(%d)\n", __FUNCTION__, real_size, inode->i_size, hidden_inode->i_size, inode->i_blocks, hidden_inode->i_blocks);

	if (real_size <= 0)
		memset(page_data, 0, PAGE_CACHE_SIZE);
	else if (real_size < PAGE_CACHE_SIZE) {
		memcpy(page_data, hidden_page_data, real_size);
		memset(page_data + real_size, 0, PAGE_CACHE_SIZE - real_size);
	} else {
		memcpy(page_data, hidden_page_data, PAGE_CACHE_SIZE);
	}

	kunmap(CURRENT_HIDDEN_PAGE);

out_release:
	if (!IS_ERR(CURRENT_HIDDEN_PAGE))
		page_cache_release(CURRENT_HIDDEN_PAGE); /* undo read_cache_page */
	print_exit_status(err);
	return err;
}






STATIC int raif_readpage_nounlock(file_t *file, page_t *page)
{
	int err;
	inode_t *inode;
	unsigned bindex, bnum;
	unsigned hidden_page_index, page_index;
	char* page_data;

	print_entry_location();

	CALCULATE_PAGE_LOCATION_READ(page, &inode, &bnum, &page_index, &bindex, &hidden_page_index);

#ifdef RAIF_DELAY_CREATE
	if((err = raif_check_and_create_file(file, bindex)))
		goto out;

#endif

	/* map pages, get their addresses */
	page_data = (char *) kmap(page);
	err = raif_do_readpage(file, page_data, bindex, hidden_page_index);
	kunmap(page);

	if (err == 0)
		SetPageUptodate(page);
	else
		ClearPageUptodate(page);

#ifdef RAIF_DELAY_CREATE
out:
#endif
	print_exit_status(err);
	return err;
}


STATIC int raif_readpage(file_t *file, page_t *page)
{
	int err = 0;
	print_entry_location();

	err = raif_readpage_nounlock(file, page);
	/* XXX: is this unlock necessary? */
	unlock_page(page);

	print_exit_status(err);
	return err;
}


STATIC int raif_prepare_write(file_t *file, page_t *page, unsigned from, unsigned to)
{
	int err = 0;

	print_entry_location();

	kmap(page);

	/* fast path for whole page writes */
	if (from == 0 && to == PAGE_CACHE_SIZE)
		goto out;

	/* read the page to "revalidate" our data */
	/* call the helper function which doesn't unlock the page */
	if (!PageUptodate(page))
		err = raif_readpage_nounlock(file, page);

out:
	print_exit_status(err);
	return err;
}




#ifdef RAIF_DELAY_CREATE
int raif_check_and_create_file(file_t *file, int bindex)
{
	int err = 0;
	file_t * first_file = NULL, * hidden_file = NULL;
	dentry_t * dentry = file->f_dentry;
	inode_t * inode = dentry->d_inode, * old_inode = NULL;
	dentry_t * lower_dentry = NULL, * lower_dir_dentry = NULL, *old_dentry = NULL;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	struct nameidata nd;
#endif

	down(&dentry->d_parent->d_inode->i_sem); /*to avoid lookups and creates at the same time as this*/
 
	if(dtohd_index(dentry, bindex) == NULL || dtohd_index(dentry, bindex)->d_inode == NULL) {

		lower_dentry = lookup_one_len(dentry->d_name.name, dtohd_index(dentry->d_parent, bindex), dentry->d_name.len);

		if(!IS_ERR(lower_dentry) && lower_dentry->d_inode) {

			old_dentry = dtohd_index(dentry, bindex);

			dtohd_index(dentry, bindex) = lower_dentry;

			if(old_dentry)
				dput(old_dentry);

			old_inode = itohi_index(inode, bindex);

			itohi_index(inode, bindex) = igrab(lower_dentry->d_inode);

			if(old_inode)
				iput(old_inode); 
		}
		else if(!IS_ERR(lower_dentry)) {
			
			lower_dir_dentry = lock_parent(lower_dentry);
			if(IS_ERR(lower_dir_dentry)) {
				err = PTR_ERR(lower_dir_dentry);
				goto out;
			}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			err = vfs_create(lower_dentry->d_parent->d_inode, lower_dentry, file->f_dentry->d_inode->i_mode);
#else
			nd.dentry = dget(lower_dir_dentry);
			nd.mnt = mntget(stohm_index(inode->i_sb, bindex));
			nd.last = lower_dentry->d_name;
			nd.last_type = LAST_NORM;

			err = vfs_create(lower_dentry->d_parent->d_inode, lower_dentry, file->f_dentry->d_inode->i_mode, &nd);
			
			dput(lower_dir_dentry);
			mntput(nd.mnt);
#endif
			
			if(err) {
				dput(lower_dentry);
				unlock_dir(lower_dir_dentry);
				goto out;
			}

			unlock_dir(lower_dir_dentry);

			old_dentry = dtohd_index(dentry, bindex);

			dtohd_index(dentry, bindex) = lower_dentry;

			if(old_dentry)
				dput(old_dentry);

			old_inode = itohi_index(inode, bindex);

			itohi_index(inode, bindex) = igrab(lower_dentry->d_inode);

			if(old_inode)
				iput(old_inode); 

			if((first_file = ftohf_index(file, 0))) {
				dget(lower_dentry);
				mntget(first_file->f_vfsmnt);
				hidden_file = dentry_open(lower_dentry, first_file->f_vfsmnt, first_file->f_flags);
				if ( IS_ERR(hidden_file)) {
					err = PTR_ERR(hidden_file);
					goto out;
				}
				ftohf_index(file, bindex) = hidden_file;
			}
		}
		else
			err = PTR_ERR(lower_dentry);
	}

out:
	up(&dentry->d_parent->d_inode->i_sem);
	return err;
}
#endif

STATIC int raif_commit_write(file_t *file, page_t *page, unsigned from, unsigned to)
{
	int err = -ENOMEM;
	inode_t *inode;
	int bnum, bindex = 0;
	unsigned page_index, hidden_page_index;
	struct iattr new_attr;
	file_t * lower_file = NULL;

	int b;
	inode_t *lower_inode;

	print_entry_location();

	CALCULATE_PAGE_LOCATION_WRITE(page, &inode, &bnum, &page_index, &bindex, &hidden_page_index);

#ifdef RAIF_DELAY_CREATE
	if ((err = raif_check_and_create_file(file, bindex)))
		goto out;
#endif

		// we may need to update lower i_size here
		bnum = ibnum(inode);
		// i_size cannot change below us because we hold i_sem
		if (PAGE_CACHE_SIZE * page->index + to > inode->i_size)
			for (b = 0; b < bnum; b++) {
				lower_inode = itohi_index(inode, b);
				lower_file = ftohf_index(file, b);
				down(&lower_inode->i_sem);

				new_attr.ia_valid = ATTR_SIZE;
				new_attr.ia_size = calculate_inode_size(inode, b, ((loff_t)(page->index) << PAGE_CACHE_SHIFT) + to);

				if ( new_attr.ia_size > i_size_read(lower_inode) )
					notify_change(lower_file->f_dentry, &new_attr);

				up(&lower_inode->i_sem);
			}

		/*
		 * We should probably use 1 << inode->i_blkbits instead of PAGE_SIZE here.
		 * However, stackable f/s copy i_blkbits from f/s below and it can be
		 * bigger than PAGE_SIZE (e.g., 32K for NFS).  PAGE_SIZE is hardcoded in
		 * the alloc_page_buffers and makes create_empty_buffers
		 * dereference a NULL pointer.  (Is it a kernel bug?)  Therefore, we have
		 * to hardcode PAGE_SIZE either here or during the i_blkbits init.
		 */
		if (!page_has_buffers(page))
			create_empty_buffers(page, PAGE_SIZE, 0);

		err = generic_commit_write(file, page, from, to);
		kunmap(page);

#ifdef RAIF_DELAY_CREATE
out:
#endif
	print_exit_status(err);
	return err;			/* assume all is ok */
}

struct address_space_operations raif_aops =
{
	writepage:	raif_writepage,
	readpage:	raif_readpage,
	prepare_write:	raif_prepare_write,
	commit_write:   raif_commit_write
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
